const {ServiceBroker} = require('moleculer')
const ConsoleLogger = require('moleculer').Loggers.Console
const FormattedLogger = require('moleculer').Loggers.Formatted
const {AsyncLocalStorage} = require('async_hooks')

const natsOptions = {
    type: 'NATS',
    options: {
        urls: ['nats://78.47.50.119'],
        user: 'console-service',
        pass: 'fdfdsdDW2ds21EW'
    }
}

const asyncLocalStorage = new AsyncLocalStorage()
asyncLocalStorage.enterWith(new Map())

class Logger extends ConsoleLogger {
    getLogHandler(bindings) {
        const level = bindings ? this.getLogLevel(bindings.mod) : null
        if (!level) {
            return null
        }
        bindings.mod = bindings.mod ?? ''
        const levelIdx = FormattedLogger.LEVELS.indexOf(level)

        return (type, args) => {
            const typeIdx = FormattedLogger.LEVELS.indexOf(type)
            if (typeIdx > levelIdx) {
                return
            }
            const store = asyncLocalStorage.getStore()
            const log = JSON.stringify({
                type: type,
                serviceName: store.get('serviceName'),
                serviceId: store.get('serviceId'),
                requestId: store.get('requestId'),
                time: Date.now(),
                trace: args.reduce((r, v) => r += ` ${v}`, ''),
                isTracing: false
            })

            switch(type) {
                case 'fatal':
                case 'error':
                case 'warn': return
                default: return console.log(log)
            }
        }
    }
}

function wrapper(err) {
    const store = asyncLocalStorage.getStore()
    return JSON.stringify({
        type: 'fatal',
        serviceName: store.get('serviceName'),
        serviceId: store.get('serviceId'),
        requestId: store.get('requestId'),
        time: Date.now(),
        trace: err.stack,
        message: err.message,
        isTracing: false
    })
}

function traceWrapper(responseTime, operation, type, params) {
    const store = asyncLocalStorage.getStore()
    return JSON.stringify({
        type: type,
        serviceName: store.get('serviceName'),
        serviceId: store.get('serviceId'),
        requestId: store.get('requestId'),
        time: Date.now(),
        operation: operation,
        responseTime: responseTime,
        isTracing: true,
        params: params
    })
}

function isPromise(obj) {
    return !!obj && (typeof obj === 'object' || typeof obj === 'function') && typeof obj.then === 'function'
}

const logger = new Logger()

module.exports = {
    middleware: {
        name: 'Logger',
        localAction(next, action) {
            return async (ctx) => {
                const store = new Map()
                return asyncLocalStorage.run(store, async () => {
                    store
                        .set('serviceName', ctx.service.name)
                        .set('serviceId', ctx.nodeID)
                        .set('requestId', ctx.requestID)
                    const start = Date.now()
                    try {
                        const result = await next(ctx)
                        const responseTime = Date.now() - start
                        console.log(traceWrapper(responseTime, action.name, 'action', ctx.params))
                        return result
                    } catch (e) {
                        console.error(wrapper(e))
                        throw e
                    }
                })
            }
        },
        localEvent(next, event) {
            return async (ctx) => {
                const store = new Map()
                return asyncLocalStorage.run(store, async () => {
                    store
                        .set('serviceName', ctx.service.name)
                        .set('serviceId', ctx.nodeID)
                        .set('requestId', ctx.requestID)
                    const start = Date.now()
                    try {
                        const result = await next(ctx)
                        const responseTime = Date.now() - start
                        return result
                    } catch (e) {
                        console.error(wrapper(e))
                        throw e
                    }
                })
            }
        },
        localMethod(next, method) {
            return (...args) => {
                const start = Date.now()

                const result = next(...args)
                if (isPromise(result)) {
                    return result.then(data => {
                        const responseTime = Date.now() - start
                        console.log(traceWrapper(responseTime, method.name, 'method', args))
                        return data
                    })
                }
                const responseTime = Date.now() - start
                console.log(traceWrapper(responseTime, method.name, 'method', args))
                return result
            }
        },
    },
    logger
}